package com.sleepycat.je.cleaner;

import com.sleepycat.je.DatabaseDeleteOperation;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.EnvironmentStats;
import com.sleepycat.je.StatsConfig;
import com.sleepycat.je.dbi.DatabaseImpl;
import com.sleepycat.je.tree.BIN;
import com.sleepycat.je.tree.DIN;
import com.sleepycat.je.txn.LockGrantType;
import com.sleepycat.je.txn.LockResult;
import com.sleepycat.je.utilint.DbLsn;

public privileged aspect CollectCleanerStats {

	/*
	 * Cumulative counters. Updates to these counters occur in multiple threads,
	 * including FileProcessor threads, and are not synchronized. This could
	 * produce errors in counting, but avoids contention around stat updates.
	 */
	int Cleaner.nBacklogFiles = 0;

	int Cleaner.nCleanerRuns = 0;

	int Cleaner.nCleanerDeletions = 0;

	int Cleaner.nINsObsolete = 0;

	int Cleaner.nINsCleaned = 0;

	int Cleaner.nINsDead = 0;

	int Cleaner.nINsMigrated = 0;

	int Cleaner.nLNsObsolete = 0;

	int Cleaner.nLNsCleaned = 0;

	int Cleaner.nLNsDead = 0;

	int Cleaner.nLNsLocked = 0;

	int Cleaner.nLNsMigrated = 0;

	int Cleaner.nLNsMarked = 0;

	int Cleaner.nLNQueueHits = 0;

	int Cleaner.nPendingLNsProcessed = 0;

	int Cleaner.nMarkedLNsProcessed = 0;

	int Cleaner.nToBeCleanedLNsProcessed = 0;

	int Cleaner.nClusterLNsProcessed = 0;

	int Cleaner.nPendingLNsLocked = 0;

	int Cleaner.nEntriesRead = 0;

	/**
	 * Load stats.
	 */
	public void Cleaner.loadStats(StatsConfig config, EnvironmentStats stat)
			throws DatabaseException {

		stat.setCleanerBacklog(nBacklogFiles);
		stat.setNCleanerRuns(nCleanerRuns);
		stat.setNCleanerDeletions(nCleanerDeletions);
		stat.setNINsObsolete(nINsObsolete);
		stat.setNINsCleaned(nINsCleaned);
		stat.setNINsDead(nINsDead);
		stat.setNINsMigrated(nINsMigrated);
		stat.setNLNsObsolete(nLNsObsolete);
		stat.setNLNsCleaned(nLNsCleaned);
		stat.setNLNsDead(nLNsDead);
		stat.setNLNsLocked(nLNsLocked);
		stat.setNLNsMigrated(nLNsMigrated);
		stat.setNLNsMarked(nLNsMarked);
		stat.setNLNQueueHits(nLNQueueHits);
		stat.setNPendingLNsProcessed(nPendingLNsProcessed);
		stat.setNMarkedLNsProcessed(nMarkedLNsProcessed);
		stat.setNToBeCleanedLNsProcessed(nToBeCleanedLNsProcessed);
		stat.setNClusterLNsProcessed(nClusterLNsProcessed);
		stat.setNPendingLNsLocked(nPendingLNsLocked);
		stat.setNCleanerEntriesRead(nEntriesRead);

		if (config.getClear()) {
			nCleanerRuns = 0;
			nCleanerDeletions = 0;
			nINsObsolete = 0;
			nINsCleaned = 0;
			nINsDead = 0;
			nINsMigrated = 0;
			nLNsObsolete = 0;
			nLNsCleaned = 0;
			nLNsDead = 0;
			nLNsLocked = 0;
			nLNsMigrated = 0;
			nLNsMarked = 0;
			nLNQueueHits = 0;
			nPendingLNsProcessed = 0;
			nMarkedLNsProcessed = 0;
			nToBeCleanedLNsProcessed = 0;
			nClusterLNsProcessed = 0;
			nPendingLNsLocked = 0;
			nEntriesRead = 0;
		}
	}

	after(Cleaner cleaner):
		execution(* Cleaner.hook_deleteFile(..)) && target(cleaner) {

		cleaner.nCleanerDeletions++;
	}

	after(Cleaner cleaner):
		execution(* Cleaner.updateReadOnlyFileCollections(..)) && target(cleaner) {

		cleaner.nBacklogFiles = cleaner.fileSelector.getBacklog();
	}

	before(Cleaner cleaner):
		execution(* Cleaner.processPendingLN(..)) && target(cleaner) {

		cleaner.nPendingLNsProcessed++;
	}

	after(Cleaner cleaner):
		execution(void hook_nLNsDead(Cleaner)) && args(cleaner) 
	{
		cleaner.nLNsDead++;
	}

	after(Cleaner cleaner) returning (LockResult lockRet):
		call(* *.nonBlockingLock(..)) && this(cleaner) && withincode(* Cleaner.processPendingLN(..)) {

		if (lockRet.getLockGrant() == LockGrantType.DENIED) {
			cleaner.nPendingLNsLocked++;
		}
	}

	after(Cleaner cleaner) returning (boolean parentFound):
		call(* *.getParentBINForChildLN(..)) && this(cleaner) && withincode(* Cleaner.processPendingLN(..)) {

		if (!parentFound) {
			cleaner.nLNsDead++;
		}
	}

	after(Cleaner cleaner, boolean migrateFlag, boolean isResident,
			boolean proactiveMigration, boolean isBinInDupDb, long childLsn):
		execution(* Cleaner.shouldMigrateLN(..)) && target(cleaner) && within(Cleaner) &&
		args(migrateFlag, isResident, proactiveMigration, isBinInDupDb, childLsn) {

		if (migrateFlag) {
			cleaner.nMarkedLNsProcessed++;

		} else if (!proactiveMigration || isBinInDupDb
				|| cleaner.env.isClosing()) {

		} else {

			Long fileNum = new Long(DbLsn.getFileNumber(childLsn));

			if ((Cleaner.PROACTIVE_MIGRATION || isResident)
					&& cleaner.mustBeCleanedFiles.contains(fileNum)) {
				cleaner.nToBeCleanedLNsProcessed++;
			} else if ((cleaner.clusterAll || (cleaner.clusterResident && isResident))
					&& cleaner.lowUtilizationFiles.contains(fileNum)) {
				cleaner.nClusterLNsProcessed++;
			}
		}
	}

	after(Cleaner cleaner):
		execution(* Cleaner.hook_nLNsLocked(..)) && target(cleaner) {

		cleaner.nLNsLocked++;
	}

	after(Cleaner cleaner):
		execution(* Cleaner.hook_nLNsDead(..)) && target(cleaner) {

		cleaner.nLNsDead++;
	}

	after(Cleaner cleaner):
		(call(void DIN.updateDupCountLNRef(long)) || 
				call(void BIN.updateEntry(int,long)) )
			&& this(cleaner)
			{

		cleaner.nLNsMigrated++;
	}

}
